import Foundation
import WebSocketKit
import CloudKit

private let encoder = JSONEncoder()

/**
 Specifies the number of cards of the called suit that a player must have
 to be allowed to play any card of the suit instead of having to play the ace.
 */
private let numberOfCardsToProtectAce = 4

let numberOfCardsPerPlayer = 8


final class Player {
    
    let name: PlayerName
    
    /// The player is the first to play a card in a new game
    var playsFirstCard = false
    
    /// The player is the next to perform an action (e.g. play a card)
    var isNextActor = false
    
    /// The player must select the game to play after winning the auction
    var selectsGame = false
    
    /// The players plays/played the first card for the current trick
    var startedCurrentTrick = false
    
    /// The action available to the player
    var actions: [Action] = []
    
    /// Indicates if the player doubled ("legen")
    var didDoubleAfterFourCards: Bool? = nil
    
    /// Indicates if the player is still involved in the bidding process
    var isStillBidding = true
    
    /// Indicates that the player leads the game ("Spieler")
    var isGameLeader = false
    
    /// Indicates the number of raises ("Schuss") of the player
    var numberOfRaises = 0
    
    /// The remaining cards of the player
    var handCards: [PlayableCard] = []
    
    /// The card played for the current trick
    var playedCard: Card? = nil
    
    /// All tricks won by the player in this game
    var wonTricks: [Trick] = []
    
    var socket: WebSocket? = nil

    var canOfferWedding: Bool {
        rawCards.canOfferWedding
    }

    var offersWedding = false

    var wouldAcceptWedding = false
    
    init(name: PlayerName) {
        self.name = name
    }
    
    var rawCards: [Card] {
        handCards.map { $0.card }
    }

    func has(card: Card) -> Bool {
        handCards.contains { $0.card == card }
    }

    func hasPlayable(card: Card) -> Bool {
        handCards.contains { $0.card == card && $0.isPlayable }
    }

    func remove(card: Card) {
        handCards = handCards.filter { $0.card != card }
    }

    func play(card: Card) {
        remove(card: card)
        playedCard = card
    }
    func connect(using socket: WebSocket) {
        _ = self.socket?.close()
        self.socket = socket
    }
    
    func send(_ info: TableInfo) {
        try? socket?.send(encodeJSON(info))
    }
    
    func disconnect() -> Bool {
        guard let socket = socket else {
            return false
        }
        do {
            try socket.close().wait()
        } catch {
            print("Failed to close socket for player: \(name): \(error)")
        }
        self.socket = nil
        return true
    }

    func canPerform(_ action: Action) -> Bool {
        actions.contains(action)
    }
    
    func prepareForFirstGame(isFirstPlayer: Bool) {
        playsFirstCard = isFirstPlayer
        isNextActor = isFirstPlayer
        startedCurrentTrick = isFirstPlayer
        actions = [.deal]
        handCards = []
        playedCard = nil
        wonTricks = []
    }
    
    func assignFirstCards(_ cards: Hand) {
        actions = [.initialDoubleCost, .noDoubleCost]
        handCards = cards.map { .init(card: $0, isPlayable: false) }
    }
    
    func didDouble(_ double: Bool) {
        actions = []
        didDoubleAfterFourCards = double
    }
    
    func assignRemainingCards(_ cards: Hand) {
        handCards = (rawCards + cards)
            .sortedCards(order: NormalCardOrder.self)
            .map { .init(card: $0, isPlayable: false) }
    }
    
    func startAuction() {
        if playsFirstCard {
            actions = [.withdrawFromAuction, .increaseOrMatchGame]
        } else {
            actions = []
        }
        if canOfferWedding {
            actions.append(.offerWedding)
        }
    }

    func offerWedding() {
        offersWedding = true
        actions = []
    }

    func weddingOfferExists() {
        guard isStillBidding else {
            return
        }
        actions = [.increaseOrMatchGame, .withdrawFromAuction]
    }

    func hasWeddingOffer() {
        guard isStillBidding else {
            return
        }
        actions = [.acceptWedding, .increaseOrMatchGame, .withdrawFromAuction]
    }

    func weddingOutbid() {
        isNextActor = false
        guard isStillBidding else {
            return
        }
        actions = []
        if offersWedding {
            offersWedding = false
            isStillBidding = false
        }
    }

    func didPerformBid() {
        isNextActor = false
        actions = []
    }

    func requiresBid() {
        isNextActor = true
        actions = [.increaseOrMatchGame, .withdrawFromAuction]
    }

    func acceptWedding() {
        wouldAcceptWedding = true
        actions = []
    }

    func weddingAccepted() {
        guard isStillBidding else {
            actions = []
            return
        }
        actions = [.increaseOrMatchGame, .withdrawFromAuction]
    }

    func auctionEnded() {
        actions = []
        isStillBidding = false
        isNextActor = false
    }

    func mustSelectWeddingCard() {
        isNextActor = true
        // Only cards which are not trump can be given to the other player
        handCards = handCards.map {
            let card = $0.card
            return .init(card: card, isPlayable: !card.isTrump(in: .hochzeit))
        }
        // Hochzeit costs double
        numberOfRaises += 1
    }

    func mustSelectGame() {
        isNextActor = true
    }

    func replace(card: Card, with other: Card) {
        remove(card: card)
        handCards.append(.init(card: other, isPlayable: false))
    }

    func replaceWeddingCard(with card: Card) -> Card {
        let index = handCards.firstIndex { $0.card.isTrump(in: .hochzeit) }!
        let removed = handCards.remove(at: index).card
        handCards.append(.init(card: card, isPlayable: false))
        return removed
    }

    func gameStarts() {
        isNextActor = playsFirstCard
        startedCurrentTrick = playsFirstCard
        actions = [.doubleDuringGame]
        isGameLeader = false
    }

    func switchLeadership() {
        isGameLeader.toggle()
        if isGameLeader {
            actions = actions.filter { $0 != .doubleDuringGame }
        } else if !actions.contains(.doubleDuringGame) {
            actions.append(.doubleDuringGame)
        }
    }

    func withdrawFromBidding() {
        isStillBidding = false
        actions = []
    }

    func didFinishTrick(canDoubleInNextRound: Bool) {
        isNextActor = false
        playedCard = nil
        if canDoubleInNextRound, !isGameLeader {
            actions = [.doubleDuringGame]
        }
    }

    func didWin(trick: Trick) {
        self.wonTricks.append(trick)
        self.isNextActor = true
    }

    func setPlayableCards(forCurrentTrick trick: Trick, in game: GameType?) {
        guard let game = game, isNextActor else {
            for i in 0..<handCards.count {
                handCards[i].isPlayable = false
            }
            return
        }
        let cards = rawCards
        guard cards.count > 1 else {
            // Last card can always be played
            setAllCards(playable: true)
            return
        }
        guard let first = trick.first else {
            setPlayableCardsForStarter(game: game)
            return
        }

        let sorter = game.sortingType

        guard sorter.isTrump(first) else {
            setPlayableCardsFollowing(suit: first.suit, game: game)
            return
        }
        guard !sorter.hasTrump(in: cards) else {
            // Must follow with trump
            handCards = cards.map {
                .init(card: $0, isPlayable: sorter.isTrump($0))
            }
            return
        }
        // Can play any card if not in calling game
        guard let suit = game.calledSuit else {
            setAllCards(playable: true)
            return
        }
        // Can play any card, except the called ace
        let ace = Card(suit, .ass)
        handCards = cards.map {
            .init(card: $0, isPlayable: $0 != ace)
        }
    }

    private func setPlayableCardsFollowing(suit: Card.Suit, game: GameType) {
        let cards = rawCards
        let sorter = game.sortingType
        // No calling game, allow all cards of same suit
        let suitCards = sorter.cards(with: suit, in: cards)

        func followSuit() {
            handCards = cards.map {
                .init(card: $0, isPlayable: !sorter.isTrump($0) && $0.suit == suit)
            }
        }
        guard let called = game.calledSuit else {
            if suitCards.isEmpty {
                // Can play any card
                setAllCards(playable: true)
            } else {
                // Must follow suit
                followSuit()
            }
            return
        }
        let ace = Card(called, .ass)
        guard called == suit else {
            if suitCards.isEmpty {
                // Exclude called ace, all others allowed
                handCards = cards.map {
                    .init(card: $0, isPlayable: $0 != ace)
                }
            } else {
                // Must follow suit (called ace automatically excluded)
                followSuit()
            }
            return
        }
        // The called suit is player, must commit ace
        guard cards.contains(ace) else {
            // Must follow suit
            followSuit()
            return
        }
        // Must play ace
        handCards = cards.map { .init(card: $0, isPlayable: $0 == ace) }
    }

    private func setPlayableCardsForStarter(game: GameType) {
        guard let suit = game.calledSuit else {
            setAllCards(playable: true)
            return
        }
        let cards = rawCards
        let ace = Card(suit, .ass)
        // Check if called ace exists, to prohibit other cards of the same suit
        guard cards.contains(ace) else {
            setAllCards(playable: true)
            return
        }
        // Jodeln
        if cards.count == numberOfCardsPerPlayer,
           cards.suitCount(suit, in: game) >= numberOfCardsToProtectAce {
            setAllCards(playable: true)
            return
        }

        // Only ace allowed for the called suit
        handCards = cards.map { card in
            let notPlayable = card.suit == suit && !card.symbol.isTrumpOrAce
            return PlayableCard(card: card, isPlayable: !notPlayable)
        }
    }

    private func setAllCards(playable: Bool) {
        for i in 0..<handCards.count {
            handCards[i].isPlayable = playable
        }
    }
    
    func info(masked: Bool) -> PlayerInfo {
        .init(player: self, isMasked: masked)
    }
}

extension Player {
    
    /// Indicate that the player is connected when at a table
    var isConnected: Bool {
        guard let socket = socket else {
            return false
        }
        guard !socket.isClosed else {
            self.socket = nil
            return false
        }
        return true
    }
}

extension Player {
    
    enum Action: String, Codable {
        /// The player can request cards to be dealt
        case deal = "deal"
        
        /// The player doubles on the initial four cards
        case initialDoubleCost = "double"
        
        /// The player does not double on the initial four cards
        case noDoubleCost = "skip"
        
        /// The player offers a wedding (one trump card)
        case offerWedding = "wedding"
        
        /// The player can choose to accept the wedding
        case acceptWedding = "accept"

        /// The player matches or increases the game during auction
        case increaseOrMatchGame = "bid"
        
        /// The player does not want to play
        case withdrawFromAuction = "out"
        
        /// The player claims to win and doubles the game cost ("schießen")
        case doubleDuringGame = "raise"

        /// The url path for the client to call (e.g. /player/deal)
        var path: String {
            rawValue
        }
    }
}

extension Player: Equatable {

    static func == (lhs: Player, rhs: Player) -> Bool {
        lhs.name == rhs.name
    }
}
