import Foundation
import WebSocketKit

private extension Int {
    
    mutating func advanceInTable() {
        self = (self + 1) % maximumPlayersPerTable
    }
}

final class Table {
    
    let id: TableId
    
    let name: TableName
    
    let isPublic: Bool
    
    var players: [Player] = []
    
    var phase: GamePhase = .waitingForPlayers
    
    var gameType: GameType? = nil
    
    var minimumPlayableGame: GameType.GameClass?
    
    /// Indicates if any player doubled during the current round, extending it to the next round
    var didDoubleInCurrentRound = false
    
    /// Indicates that all players acted after the first four cards
    var allPlayersFinishedDoubling: Bool {
        !players.contains { $0.didDoubleAfterFourCards == nil }
    }

    var weddingOfferExists: Bool {
        players.contains { $0.offersWedding }
    }

    var weddingAcceptExists: Bool {
        players.contains { $0.wouldAcceptWedding }
    }

    var hasAuctionWinner: Bool {
        numberOfRemainingBidders == 1
    }

    var numberOfRemainingBidders: Int {
        players.filter { $0.isStillBidding }.count
    }

    var auctionWinner: Player {
        players.first { $0.isStillBidding }!
    }

    var hasCompletedTrick: Bool {
        !players.contains { $0.playedCard == nil }
    }

    var completedTrick: Trick? {
        let trick = players.compactMap { $0.playedCard }
        guard trick.count == maximumPlayersPerTable else {
            return nil
        }
        return trick
    }

    var currentTrick: [Card] {
        players.compactMap { $0.playedCard }
    }
    
    init(id: TableId, name: TableName, isPublic: Bool) {
        self.id = id
        self.name = name
        self.isPublic = isPublic
    }
    
    init(newTable name: TableName, isPublic: Bool) {
        self.id = .newToken()
        self.name = name
        self.isPublic = isPublic
    }
    
    func add(player: PlayerName) -> Bool {
        guard !isFull else {
            return false
        }
        let player = Player(name: player)
        players.append(player)
        if isFull {
            prepareTableForFirstGame()
        }
        sendUpdateToAllPlayers()
        return true
    }
    
    func contains(player: PlayerName) -> Bool {
        players.contains { $0.name == player }
    }

    /// The player to play the first card of the current game
    var firstPlayer: Player {
        players.first { $0.playsFirstCard }!
    }
    
    func select(player: PlayerName) -> Player? {
        players.first { $0.name == player }
    }
    
    func player(leftOf index: Int) -> Player? {
        player(at: (index + 1) % 4)
    }
    
    func player(acrossOf index: Int) -> Player? {
        player(at: (index + 2) % 4)
    }
    
    func player(rightOf index: Int) -> Player? {
        player(at: (index + 3) % 4)
    }
    
    func player(at index: Int) -> Player? {
        guard index < players.count else {
            return nil
        }
        return players[index]
    }

    func index(of player: Player) -> Int {
        players.firstIndex(of: player)!
    }

    func nextPlayer(after player: Player) -> Player {
        let i = index(of: player)
        let newIndex = (i + 1) % maximumPlayersPerTable
        return players[newIndex]
    }

    func nextBidder(after player: Player) -> Player {
        // Find next player to place bid
        let index = index(of: player)
        for i in 1..<4 {
            let player = players[(index + i) % 4]
            guard player.isStillBidding else {
                continue
            }
            return player
        }
        return player
    }
    
    func remove(player: PlayerName) {
        guard contains(player: player) else {
            return
        }
        players = players.filter { $0.name != player }
        reset()
    }
    
    func connect(player name: PlayerName, using socket: WebSocket) -> Bool {
        guard let player = select(player: name) else {
            return false
        }
        player.connect(using: socket)
        sendUpdateToAllPlayers()
        return true
    }
    
    func disconnect(player name: PlayerName) {
        guard let player = select(player: name) else {
            return
        }
        guard player.disconnect() else {
            return
        }
        sendUpdateToAllPlayers()
        return
    }
    
    private func prepareTableForFirstGame() {
        phase = .waitingForPlayers
        gameType = nil
        minimumPlayableGame = nil // Not relevant in this phase
        didDoubleInCurrentRound = false // Not relevant in this phase
        let index = players.firstIndex { $0.playsFirstCard } ?? 0
        for i in 0..<maximumPlayersPerTable {
            players[i].prepareForFirstGame(isFirstPlayer: i == index)
        }
    }
    
    private func sendUpdateToAllPlayers() {
        players.enumerated().forEach { playerIndex, player in
            guard player.isConnected else {
                return
            }
            let info = TableInfo(self, forPlayerAt: playerIndex)
            player.send(info)
        }
    }
    
    // MARK: Player actions

    func play(card: Card, player name: PlayerName) -> PlayCardResult {
        let player = select(player: name)!
        if phase == .selectWeddingCard {
            return selectedCardForWedding(card: card, player: player)
        }
        guard let game = gameType,
                player.hasPlayable(card: card) else {
            return .invalidCard
        }
        player.play(card: card)
        if let completedTrick = completedTrick {
            didFinish(trick: completedTrick, in: game)
        } else {
            let next = nextPlayer(after: player)
            next.isNextActor = true
            player.isNextActor = false
        }
        updatePlayableCards()
        return .success
    }

    private func updatePlayableCards() {
        let playedCards = currentTrick
        players.forEach {
            $0.setPlayableCards(forCurrentTrick: playedCards, in: gameType)
        }
    }

    func didFinish(trick: Trick, in game: GameType) {
        // If trick is completed, calculate winner
        let index = trick.highCardIndex(forGame: game)
        players.forEach {
            $0.didFinishTrick(canDoubleInNextRound: didDoubleInCurrentRound)
        }
        players[index].didWin(trick: trick)
        didDoubleInCurrentRound = false
    }
    
    func perform(action: Player.Action, forPlayer player: PlayerName) -> PlayerActionResult {
        let player = select(player: player)!
        guard player.canPerform(action) else {
            return .tableStateInvalid
        }
        defer { sendUpdateToAllPlayers() }
        switch action {
        case .deal:
            return dealInitialCards()
        case .initialDoubleCost:
            return perform(double: true, forPlayer: player)
        case .noDoubleCost:
            return perform(double: false, forPlayer: player)
        case .offerWedding:
            return performWeddingCall(forPlayer: player)
        case .acceptWedding:
            return handleWeddingAccept(forPlayer: player)
        case .increaseOrMatchGame:
            return performBidIncrease(forPlayer: player)
        case .withdrawFromAuction:
            return performWithdrawl(forPlayer: player)
        case .doubleDuringGame:
            return performDoubleDuringGame(forPlayer: player)
        }
    }
    
    private func dealInitialCards() -> PlayerActionResult {
        guard isFull else {
            return .tableNotFull
        }
        guard phase == .waitingForPlayers else {
            return .tableStateInvalid
        }
        
        let cards = Dealer.dealFirstCards()
        for (index, player) in players.enumerated() {
            player.assignFirstCards(cards[index])
        }
        phase = .collectingDoubles
        gameType = nil
        return .success
    }
    
    func perform(double: Bool, forPlayer player: Player) -> PlayerActionResult {
        player.didDouble(double)
        if allPlayersFinishedDoubling {
            dealAdditionalCards()
        }
        return .success
    }
    
    private func dealAdditionalCards() {
        let cards = Dealer.dealRemainingCards(of: players.map { $0.rawCards })
        for (index, player) in players.enumerated() {
            player.assignRemainingCards(cards[index])
        }
        players.forEach { $0.startAuction() }
        minimumPlayableGame = nil
    }

    private func performWeddingCall(forPlayer player: Player) -> PlayerActionResult {
        guard phase == .bidding else {
            return .tableStateInvalid
        }
        guard minimumPlayableGame == nil || minimumPlayableGame == .ruf else {
            return .tableStateInvalid
        }
        guard player.offersWedding else {
            return .tableStateInvalid
        }
        guard !weddingOfferExists else {
            // Only one wedding allowed at the table
            return .tableStateInvalid
        }
        // Only allow wedding acceptance or outbidding
        players.forEach { $0.weddingOfferExists() }
        player.offerWedding()
        firstPlayer.hasWeddingOffer()
        minimumPlayableGame = .bettel
        return .success
    }

    private func performBidIncrease(forPlayer player: Player) -> PlayerActionResult {
        guard phase == .bidding else {
            return .tableStateInvalid
        }
        if weddingOfferExists {
            // Anyone except the offerer can outbid a wedding
            return handleWeddingOutbid(forPlayer: player)
        }
        guard player.isNextActor else {
            return .tableStateInvalid
        }
        if minimumPlayableGame == nil {
            minimumPlayableGame = .ruf
        } else {
            minimumPlayableGame!.increase()
        }
        player.didPerformBid()
        // Find next player to place bid
        nextBidder(after: player).requiresBid()
        return .success
    }

    private func handleWeddingOutbid(forPlayer player: Player) -> PlayerActionResult {
        if player.offersWedding {
            // A player offering a wedding can't outbid itself
            return .tableStateInvalid
        }
        players.forEach { $0.weddingOutbid() }
        firstPlayer.requiresBid()
        return .success
    }

    private func handleWeddingAccept(forPlayer player: Player) -> PlayerActionResult {
        guard phase == .bidding else {
            return .tableStateInvalid
        }
        guard minimumPlayableGame == nil || minimumPlayableGame == .ruf else {
            return .tableStateInvalid
        }
        guard weddingOfferExists else {
            return .tableStateInvalid
        }
        guard player.isNextActor else {
            return .tableStateInvalid
        }
        guard !player.offersWedding else {
            return .tableStateInvalid
        }
        guard !weddingAcceptExists else {
            return .tableStateInvalid
        }
        if hasAuctionWinner {
            selectedWedding(player: player)
            return .success
        }
        minimumPlayableGame = .bettel
        players.forEach {
            $0.weddingAccepted()
        }
        player.acceptWedding()
        nextBidder(after: player).requiresBid()
        return .success
    }

    private func selectedWedding(player: Player) {
        minimumPlayableGame = nil
        gameType = .hochzeit
        phase = .selectWeddingCard
        players.forEach { $0.auctionEnded() }
        player.mustSelectWeddingCard()
    }

    private func selectedCardForWedding(card: Card, player: Player) -> PlayCardResult {
        guard player.isNextActor,
              player.wouldAcceptWedding,
              weddingOfferExists else {
                  return .invalidTableState
              }
        guard !card.isTrump(in: .hochzeit),
              player.has(card: card) else {
                  return .invalidCard
              }
        // Swap the cards
        let offerer = players.first { $0.offersWedding }!
        let offeredCard = offerer.replaceWeddingCard(with: card)
        player.replace(card: card, with: offeredCard)

        // Start the game
        gameType = .hochzeit
        players.forEach { $0.gameStarts() }
        player.switchLeadership()
        offerer.switchLeadership()
        return .success
    }

    private func performWithdrawl(forPlayer player: Player) -> PlayerActionResult {
        guard phase == .bidding,
              player.isNextActor,
              player.isStillBidding else {
                  return .tableStateInvalid
              }
        player.withdrawFromBidding()
        switch numberOfRemainingBidders {
        case 1:
            selectGame(player: auctionWinner)
        case 0:
            // All players withdrawn, deal new cards
            let first = firstPlayer
            let newPlayer = self.nextBidder(after: first)
            first.playsFirstCard = false
            newPlayer.playsFirstCard = true
            prepareTableForFirstGame()
            return dealInitialCards()
        default:
            break
        }
        return .success
    }

    private func selectGame(player: Player) {
        minimumPlayableGame = nil
        gameType = nil
        phase = .selectGame
        players.forEach { $0.auctionEnded() }
        player.mustSelectGame()
    }

    private func performDoubleDuringGame(forPlayer player: Player) -> PlayerActionResult {
        guard phase == .playing,
              !player.isGameLeader else {
                  return .tableStateInvalid
              }
        player.numberOfRaises += 1
        players.forEach { $0.switchLeadership() }
        return .success
    }

    private func reset() {
        phase = .waitingForPlayers
        gameType = nil
        minimumPlayableGame = nil
    }
}

extension Table {
    
    var isFull: Bool {
        players.count == maximumPlayersPerTable
    }
    
    var publicInfo: PublicTableInfo {
        .init(id: id, name: name, players: playerNames)
    }
    
    var playerNames: [PlayerName] {
        players.map { $0.name }
    }
    
    func compileInfo(for player: PlayerName) -> TableInfo? {
        guard let index = players.firstIndex(where: { $0.name == player }) else {
            return nil
        }
        
        return TableInfo(self, forPlayerAt: index)
    }
}
